# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'bilby NR'
copyright = '2026, C.Hoy'
author = 'C.Hoy'
release = '0.1.0'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = ["sphinx.ext.ifconfig", "sphinx_immaterial", "sphinx_tabs.tabs"]

templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']



# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'sphinx_immaterial'
html_theme_options = {
    # colouring and light/dark mode
    "palette": [
        {
            "media": "(prefers-color-scheme: light)",
            "scheme": "default",
            "primary": "blue-grey",
            "accent": "deep-orange",
        },
    ]
}
html_static_path = ['_static']



from bilby_nr import utils
import pytest

def test_convert_waveform_input_ValueError():
    with pytest.raises(ValueError):
        utils.convert_waveform_input(None)


def test_convert_waveform_input_list():
    out = utils.convert_waveform_input(["A", "B", "C"])
    assert out == ["A", "B", "C"]


def test_convert_waveform_input_string():
    variation = [
        "[A, B, C]",
        "['A', 'B', 'C']",
        "'A', 'B', 'C'",
        "A, B, C",
        "A,B,C",
    ]
    for opt in variation:
        out = utils.convert_waveform_input(opt)
        assert out == ["A", "B", "C"]



# Licensed under an MIT style license -- see LICENSE.md

import copy
import numpy as np
from bilby.gw import conversion
from bilby.gw.conversion import _generate_all_cbc_parameters as _base_generate

__author__ = ["Charlie Hoy <charlie.hoy@port.ac.uk>"]


def chi_par_chi_perp_from_mass_spin(
    mass_1, mass_2, a_1, tilt_1, a_2, tilt_2, phi_12, phi_jl, theta_jn, phase,
    reference_frequency
):
    """Calculate and return the parallel and perpendicular spin, as defined in
    Eqs. 11 and 12 of https://www.nature.com/articles/s41550-025-02579-7

    Parameters
    ----------
    mass_1: float
        The mass of the primary black hole
    mass_2: float
        The mass of the secondary black hole
    a_1: float
        The dimensionless spin magnitude of the primary black hole
    tilt_1: float
        The tilt angle of the primary black hole spin
    a_2: float
        The dimensionless spin magnitude of the secondary black hole
    tilt_2: float
        The tilt angle of the secondary black hole spin
    phi_12: float
        The difference in azimuthal angle between the two spins
    phi_jl: float
        The azimuthal angle of the total angular momentum
    theta_jn: float
        The angle between the total angular momentum and the line of sight
    phase: float
        The phase of the gravitational wave
    reference_frequency: float
        The reference frequency

    Returns
    -------
    chi_par: float
        The parallel spin
    chi_perp: float
        The perpendicular spin
    """
    _chi_par = chi_par(mass_1, mass_2, a_1, tilt_1, a_2, tilt_2)
    _chi_perp = chi_perp(
        mass_1, mass_2, a_1, tilt_1, a_2, tilt_2, phi_12, phi_jl, theta_jn,
        phase, reference_frequency
    )
    return _chi_par, _chi_perp


def chi_par(mass_1, mass_2, a_1, tilt_1, a_2, tilt_2):
    """Calculate and return the parallel spin, as defined in Eq.12 of
    https://www.nature.com/articles/s41550-025-02579-7

    Parameters
    ----------
    mass_1: float
        The mass of the primary black hole
    mass_2: float
        The mass of the secondary black hole
    a_1: float
        The dimensionless spin magnitude of the primary black hole
    tilt_1: float
        The tilt angle of the primary black hole spin
    a_2: float
        The dimensionless spin magnitude of the secondary black hole
    tilt_2: float
        The tilt angle of the secondary black hole spin

    Returns
    -------
    chi_par: float
        The parallel spin
    """
    mass_ratio = mass_2 / mass_1
    numerator = (
        a_1 * np.cos(tilt_1) + mass_ratio**2 * a_2 * np.cos(tilt_2)
    )
    denominator = (1 + mass_ratio)**2
    return numerator / denominator


def chi_perp(
    mass_1, mass_2, a_1, tilt_1, a_2, tilt_2, phi_12, phi_jl, theta_jn, phase,
    reference_frequency
):
    """Calculate and return the perpependicular spin, as defined in Eq. 11 of
    https://www.nature.com/articles/s41550-025-02579-7

    Parameters
    ----------
    mass_1: float
        The mass of the primary black hole
    mass_2: float
        The mass of the secondary black hole
    a_1: float
        The dimensionless spin magnitude of the primary black hole
    tilt_1: float
        The tilt angle of the primary black hole spin
    a_2: float
        The dimensionless spin magnitude of the secondary black hole
    tilt_2: float
        The tilt angle of the secondary black hole spin
    phi_12: float
        The difference in azimuthal angle between the two spins
    phi_jl: float
        The azimuthal angle of the total angular momentum
    theta_jn: float
        The angle between the total angular momentum and the line of sight
    phase: float
        The phase of the gravitational wave
    reference_frequency: float
        The reference frequency

    Returns
    -------
    chi_perp: float
        The perpendicular spin
    """
    total_mass = mass_1 + mass_2
    spins = conversion.bilby_to_lalsimulation_spins(
        theta_jn, phi_jl, tilt_1, tilt_2, phi_12, a_1, a_2, mass_1, mass_2,
        reference_frequency, phase
    )
    S1_perp = mass_1**2 * np.array([spins[1], spins[2]])
    S2_perp = mass_2**2 * np.array([spins[4], spins[5]])
    S_perp = np.sum([S1_perp, S2_perp], axis=0)
    S_perp_mag = np.linalg.norm(S_perp)
    chi_perp = S_perp_mag / total_mass**2
    return chi_perp


def _generate_all_cbc_parameters(
    sample, defaults, base_conversion, likelihood=None, priors=None, npool=1
):
    """Extension of bilby.gw.conversion.generate_all_bbh_parameters to allow
    for multiple models

    Parameters
    ==========
    sample: dict, pandas.DataFrame
        Samples to fill in with extra parameters, this may be either an
        injection or posterior samples.
    likelihood: bilby.gw.likelihood.GravitationalWaveTransient, optional
        GravitationalWaveTransient used for sampling, used for waveform and
        likelihood.interferometers.
    priors: dict, optional
        Dictionary of prior objects, used to fill in non-sampled parameters.
    npool: int, optional
        Number of processes to use for the conversion. Default 1
    """
    _chosen_model = None
    if "log_likelihood" in sample.keys() and likelihood is not None:
        waveform_approximant_list = \
            likelihood.waveform_generator.waveform_arguments.get(
                "waveform_approximant_list", None
            )
        if waveform_approximant_list is not None:
            _chosen_model = determine_waveform_approximant_from_likelihood(
                sample, waveform_approximant_list, likelihood
            )

    if _chosen_model is None:
        # assume a default
        _chosen_model = "IMRPhenomTPHM"
    _likelihood = copy.deepcopy(likelihood)
    if _chosen_model is not None:
        if likelihood is not None:
            _likelihood.waveform_generator.waveform_arguments.update(
                {
                    "waveform_approximant_list": [_chosen_model],
                    "waveform_approximant": _chosen_model
                }
            )
        sample["waveform_approximant"] = _chosen_model
        defaults["waveform_approximant"] = _chosen_model
    return _base_generate(
        sample, defaults=defaults,
        base_conversion=base_conversion,
        likelihood=_likelihood, priors=priors, npool=npool
    )


def determine_waveform_approximant_from_likelihood(
    sample, waveform_approximant_list, likelihood
):
    """Determine the waveform approximant that evaluated the likelihood during
    the inference.

    Parameters
    ----------
    sample: dict, pandas.DataFrame
        the sample you wish to evaluate the likelihood for
    waveform_approximant_list: list
        list of waveform approximants you wish to consider
    likelihood: bilby.gw.likelihood.GravitationalWaveTransient
        likelihood object that was used during the sampling
    """
    _chosen_model = None
    _sample = sample.copy()
    if "log_likelihood" not in _sample.keys():
        raise ValueError(
            "Unable to determine which model ws used to evaluate the "
            "likelihood as there is no 'log_likelihood' entry in the provided "
            "samples."
        )
    original = _sample.pop("log_likelihood")
    if not isinstance(original, (float, int, np.number)):
        if isinstance(_sample, dict):
            original = original[0]
        else:
            original = original.values[0]
            _sample = _sample.to_dict(orient="list")
            _sample = {key: item[0] for key, item in _sample.items()}

    for model in waveform_approximant_list:
        _likelihood = copy.deepcopy(likelihood)
        logl = _likelihood_for_given_model(_sample, model, _likelihood)
        if np.isclose(logl, original):
            _chosen_model = model
            break
    if _chosen_model is None:
        raise ValueError(
           "Unable to find a model that returns the same log likelihood as "
           "that stored in the sample."
        )
    return _chosen_model


def _likelihood_for_given_model(sample, waveform_approximant, likelihood):
    """Evaluate the likelihood for a given waveform approximant

    Parameters
    ----------
    sample: dict
        the sample you wish to evaluate the likelihood for
    waveform_approximant: str
        the waveform approximant you wish to evaluate the likelihood for
    likelihood: bilby.gw.likelihood.GravitationalWaveTransient
        likelihood object that was used during the sampling
    """
    _lkl = copy.deepcopy(likelihood)
    # remove cache of last sample by adding a small shift to all parameters
    _sample = sample.copy()
    for key in _sample.keys():
        _sample[key] += np.random.random() * 1e-5
    _lkl.log_likelihood_ratio(_sample)
    # now check the sample provided
    _lkl.waveform_generator.waveform_arguments["waveform_approximant_list"] = [
        waveform_approximant
    ]
    _lkl.waveform_generator.waveform_arguments["waveform_approximant"] = \
        waveform_approximant
    _sample = {
        key: item for key, item in sample.items() if key not in
        ["log_likelihood", "log_prior"]
    }
    if not isinstance(list(_sample.values())[0], (float, int, np.number)):
        _sample = {key: item[0] for key, item in _sample.items()}
            
    if _lkl.distance_marginalization:
        _sample["luminosity_distance"] = _lkl._ref_dist
    if _lkl.time_marginalization:
        _sample["geocent_time"] = _lkl.interferometers.start_time
        if "time_jitter" not in _sample.keys():
            raise ValueError(
                "Please provide a jitter time as time marginalization was "
                "used during the sampling"
            )
    if _lkl.phase_marginalization:
        _sample["phase"] = 0.

    logl = _lkl.log_likelihood_ratio(parameters=_sample)
    return logl


conversion._generate_all_cbc_parameters = _generate_all_cbc_parameters



# Licensed under an MIT style license -- see LICENSE.md

import numpy as np
from bilby.gw import source
from bilby.core.utils import logger
import ast
from .utils import convert_waveform_list_from_input

__author__ = ["Charlie Hoy <charlie.hoy@port.ac.uk>"]


def multi_model_binary_black_hole(
    frequency_array, mass_1, mass_2, luminosity_distance, a_1, tilt_1,
    phi_12, a_2, tilt_2, phi_jl, theta_jn, phase, **kwargs
):
    """Source model for a binary black hole with multiple models.

    Parameters
    ----------
    frequency_array: np.ndarray
        The frequency array
    mass_1: float
        The mass of the primary black hole
    mass_2: float
        The mass of the secondary black hole
    luminosity_distance: float
        The luminosity distance
    a_1: float
        The dimensionless spin magnitude of the primary black hole
    tilt_1: float
        The tilt angle of the primary black hole spin
    phi_12: float
        The difference in azimuthal angle between the two spins
    a_2: float
        The dimensionless spin magnitude of the secondary black hole
    tilt_2: float
        The tilt angle of the secondary black hole spin
    phi_jl: float
        The azimuthal angle of the total angular momentum
    theta_jn: float
        The angle between the total angular momentum and the line of sight
    phase: float
        The phase of the gravitational wave
    kwargs: dict
        Additional keyword arguments. We support additional arguments beyond
        those allowed in bilby. For example:

            - waveform_approximant_list: a list of waveform approximants
              you wish to sample over. If the user provides the waveform
              approximant 'IMRPhenomXPHMST', we populate the required flags
              to use the Spin-Taylor (ST) variant of IMRPhenomXPHM.
            - match_interpolant: the interpolant you wish to use to estimate the
              match for a given region in the parameter space
            - use_best_match: always use the model with the best match to
              evaluate the likelihood
            - match_to_weight: a string that can be evaluated to map an array
              of matches to a series of weights. The model will then be
              chosen probabilistically based on the weights.

    Returns
    -------
    polarizations: dict
        The polarizations
    """
    waveform_approximant_list = kwargs.pop("waveform_approximant_list", None)
    if waveform_approximant_list is None:
        raise ValueError(
            "Please provide a list of waveforms to sample over via the "
            "waveform_approximant_list waveform argument"
        )
    waveform_approximant_list = convert_waveform_list_from_input(
        waveform_approximant_list
    )
    match_interpolant = kwargs.pop("match_interpolant", None)
    catch_waveform_errors = kwargs.get("catch_waveform_errors", False)
    try:
        if match_interpolant is not None:
            return _multi_model_match_informed_binary_black_hole(
                match_interpolant, waveform_approximant_list, frequency_array,
                mass_1, mass_2, luminosity_distance, a_1, tilt_1, phi_12, a_2,
                tilt_2, phi_jl, theta_jn, phase, **kwargs
            )
        return _multi_model_binary_black_hole(
            np.ones(len(waveform_approximant_list)) / len(waveform_approximant_list),
            waveform_approximant_list, frequency_array, mass_1, mass_2,
            luminosity_distance, a_1, tilt_1, phi_12, a_2, tilt_2, phi_jl, theta_jn,
            phase, **kwargs
        )
    except Exception as e:
        if not catch_waveform_errors:
            raise

        failed_parameters = dict(
            mass_1=mass_1, mass_2=mass_2,
            a_1=a_1, tilt_1=tilt_1, phi_12=phi_12,
            a_2=a_2, tilt_2=tilt_2, phi_jl=phi_jl,
            luminosity_distance=luminosity_distance,
            theta_jn=theta_jn, phase=phase,
        )
        logger.warning(
            "Evaluating the waveform failed with error: {}\n".format(e) +
            "The parameters were {}\n".format(failed_parameters) +
            "Likelihood will be set to -inf."
        )
        return None


def _multi_model_match_informed_binary_black_hole(
    interpolant, waveform_approximant_list, frequency_array, mass_1, mass_2,
    luminosity_distance, a_1, tilt_1, phi_12, a_2, tilt_2, phi_jl, theta_jn,
    phase, **kwargs
):
    """Source model for a binary black hole with multiple models. The model
    used to generate the GW polarizations is chosen based on the match to
    numerical relativity simulations. The match is calculated based on the
    specified interpolant

    Parameters
    ----------
    interpolant: str
        The interpolant you wish to use to estimate the match
    waveform_approximant_list: list
        list of waveform approximants you wish to include
    frequency_array: np.ndarray
        The frequency array
    mass_1: float
        The mass of the primary black hole
    mass_2: float
        The mass of the secondary black hole
    luminosity_distance: float
        The luminosity distance
    a_1: float
        The dimensionless spin magnitude of the primary black hole
    tilt_1: float
        The tilt angle of the primary black hole spin
    phi_12: float
        The difference in azimuthal angle between the two spins
    a_2: float
        The dimensionless spin magnitude of the secondary black hole
    tilt_2: float
        The tilt angle of the secondary black hole spin
    phi_jl: float
        The azimuthal angle of the total angular momentum
    theta_jn: float
        The angle between the total angular momentum and the line of sight
    phase: float
        The phase of the gravitational wave
    kwargs: dict
        Additional keyword arguments

    Returns
    -------
    polarizations: dict
        The polarizations
    """
    import importlib
    try:
        _split = interpolant.split(".")
        _module = ".".join(_split[:-1])
        _function = _split[-1]
        module = importlib.import_module(_module)
        method = getattr(module, _function)
    except Exception as e:
        raise ValueError(f"Unable to import interpolant function because: {e}")

    _matches = np.array(
        [
            method(
                wvf, mass_1, mass_2, a_1, tilt_1, phi_12, a_2, tilt_2, phi_jl,
                theta_jn, phase, 
            ) for wvf in waveform_approximant_list
        ]
    )
    # protect against negative matches
    _matches[_matches < 0.] = 0.
    use_best = kwargs.pop("use_best_match", False)
    mapping = kwargs.pop("match_to_weight", None)
    if isinstance(use_best, str):
        use_best = ast.literal_eval(use_best)
    _weights = _weights_from_matches(_matches, use_best=use_best, mapping=mapping)
    return _multi_model_binary_black_hole(
        _weights, waveform_approximant_list, frequency_array, mass_1, mass_2,
        luminosity_distance, a_1, tilt_1, phi_12, a_2, tilt_2, phi_jl,
        theta_jn, phase, **kwargs
    )


def _multi_model_binary_black_hole(
    weights, waveform_approximant_list, frequency_array, mass_1, mass_2,
    luminosity_distance, a_1, tilt_1, phi_12, a_2, tilt_2, phi_jl, theta_jn,
    phase, **kwargs
):
    """Source model for a binary black hole with multiple models. The model
    is chosen based on the weights provided.

    Parameters
    ----------
    weights: list
        The interpolant you wish to use to estimate the match
    waveform_approximant_list: list
        list of waveform approximants you wish to include. Must equal the length
        of weights
    frequency_array: np.ndarray
        The frequency array
    mass_1: float
        The mass of the primary black hole
    mass_2: float
        The mass of the secondary black hole
    luminosity_distance: float
        The luminosity distance
    a_1: float
        The dimensionless spin magnitude of the primary black hole
    tilt_1: float
        The tilt angle of the primary black hole spin
    phi_12: float
        The difference in azimuthal angle between the two spins
    a_2: float
        The dimensionless spin magnitude of the secondary black hole
    tilt_2: float
        The tilt angle of the secondary black hole spin
    phi_jl: float
        The azimuthal angle of the total angular momentum
    theta_jn: float
        The angle between the total angular momentum and the line of sight
    phase: float
        The phase of the gravitational wave
    kwargs: dict
        Additional keyword arguments

    Returns
    -------
    polarizations: dict
        The polarizations
    """
    if not np.any(weights):
        raise ValueError(
            "Input domain error. All weights are non-numeric. Please provide a "
            "numeric weight for each approximant in the list"
        )
    elif len(weights) != len(waveform_approximant_list):
        raise ValueError(
            "Please provide a weight for each approximant in the list"
        )
    else:
        waveform_approximant = np.random.choice(
            waveform_approximant_list, p=weights
        )
    kwargs["waveform_approximant"] = waveform_approximant
    # only use gwsignal for reviewed waveforms. This should be changed when
    # bilby updates their review statement
    if waveform_approximant in ["SEOBNRv5HM", "SEOBNRv5PHM"]:
        return source.gwsignal_binary_black_hole(
            frequency_array, mass_1, mass_2, luminosity_distance, a_1, tilt_1,
            phi_12, a_2, tilt_2, phi_jl, theta_jn, phase, **kwargs
        )
    if waveform_approximant == "IMRPhenomXPHMST":
        kwargs.update(
            {
                "waveform_approximant": "IMRPhenomXPHM",
                "PhenomXPrecVersion": 320,
                "PhenomXPFinalSpinMod": 2,
                "PhenomXHMReleaseVersion": 122022
            }
        )
    return source.lal_binary_black_hole(
        frequency_array, mass_1, mass_2, luminosity_distance, a_1, tilt_1,
        phi_12, a_2, tilt_2, phi_jl, theta_jn, phase, **kwargs
    )


def _weights_from_matches(matches, use_best=False, mapping=None):
    """Calculate a weight based on the match to numerical relativity. If
    mapping is None, we use the recommendation from
    https://www.nature.com/articles/s41550-025-02579-7,
    i.e. weight = 1 / (1 - match)**4

    Parameters
    ----------
    matches: np.ndarray
        array of matches to calculate the weight for
    use_best: bool, optional
        if True, return a weight of 1 for the highest match and 0 for all other
        models. Default False
    mapping: str, optional
        a string that can be evaluated to map an array of matches to a series of
        weights. The string must contain the variable 'matches' and should
        contain the right hand side of the equation 'weights = f(matches)'.
        For example, you could provide '1 / ((1 - matches)**4)'. The weights
        will be rescaled to be between 0 and 1.
    """
    if use_best:
        weights = np.zeros(len(matches))
        weights[np.argmax(matches)] = 1.
        return weights
    if mapping is None:
        weights = 1 / ((1 - matches)**4)
    else:
        if "matches" not in mapping:
            raise ValueError(
                f"{mapping} must be a string that can be evaluated and contain "
                f"a variable called: 'matches'"
            )
        if "=" in mapping:
            raise ValueError(
                f"{mapping} must not contain '='. It should contain the "
                f"right hand side of the equation 'weights = f(matches)'. For "
                f"example, you could provide '1 / ((1 - matches)**4)'"
            )
        try:
            weights = eval(mapping)
        except Exception as e:
            raise ValueError(
                f"Unable to generate weights from matches for the string "
                f"{mapping}."
            )
    weights /= np.sum(weights)
    weights[np.isnan(weights)] = 0.
    return weights



from bilby_nr.bilby_pipe import Input
from bilby.gw.detector.networks import InterferometerList
from bilby_pipe.utils import BilbyPipeError
import os
import pytest
import tempfile

tmpdir = tempfile.TemporaryDirectory(prefix=".", dir=".").name


class TestInput(object):
    def test_original_waveform_approximant_behaviour(self):
        inputs = Input(None, None)
        inputs.waveform_approximant = "A"
        assert inputs.waveform_approximant == "A"

    def test_waveform_approximant_list_with_correct_source_model(self):
        variations = [
            ["A", "B"], "[A, B]", "['A', 'B']", "A, B"
        ]
        for opt in variations:
            inputs = Input(None, None)
            inputs.frequency_domain_source_model = (
                "bilby_nr.source.multi_model_binary_black_hole"
            )
            inputs.waveform_approximant = opt
            assert inputs.waveform_approximant == ["A", "B"]

    def test_waveform_approximant_list_with_incorrect_source_model(self):
        inputs = Input(None, None)
        inputs.frequency_domain_source_model = "lal_binary_black_hole"
        with pytest.raises(BilbyPipeError):
            inputs.waveform_approximant = ["A", "B"]

    def test_waveform_generator(self):
        for num, approx in enumerate(["A", "[A, B]"]):
            inputs = Input(None, None)
            inputs.detectors = ["H1"]
            inputs.interferometers = InterferometerList(["H1"])
            inputs.conversion_function = "noconvert"
            inputs.waveform_generator_class = "WaveformGenerator"
            inputs.detectors = ["H1"]
            inputs.waveform_generator_class_ctor_args = None
            inputs.likelihood_type = "GravitationalWaveTransient"
            if num == 0:
                inputs.frequency_domain_source_model = "lal_binary_black_hole"
            else:
                inputs.frequency_domain_source_model = (
                    "bilby_nr.source.multi_model_binary_black_hole"
                )
            inputs.waveform_approximant = approx
            inputs.catch_waveform_errors = False
            inputs.reference_frequency = 20
            inputs.minimum_frequency = 20
            inputs.maximum_frequency = 1024
            inputs.pn_spin_order = -1
            inputs.pn_tidal_order = -1
            inputs.pn_phase_order = -1
            inputs.pn_amplitude_order = 0
            inputs.mode_array = None
            inputs.waveform_arguments_dict = None
            args = inputs.waveform_generator.waveform_arguments
            assert args['waveform_approximant'] == "A"
            if num == 0:
                assert 'waveform_approximant_list' not in args.keys()
            else:
                assert args['waveform_approximant_list'] == ["A", "B"]


def TestMainInput(object):
    def setup_method(self):
        from bilby_nr.bilby_pipe import (
            create_parser, MainInput
        )
        from bilby_pipe.main import (
            create_parser as original_parser,
            MainInput as OriginalInput
        )
        from bilby_pipe.main import parse_args

        self.outdir = tmpdir
        self.default_args_list = [
            "--ini",
            os.path.join(os.path.dirname(__file__), "test_config.ini"),
            "--outdir",
            self.outdir,
        ]
        self.my_parser = create_parser()
        self.my_inputs = MainInput(
            *parse_args(self.default_args_list, self.my_parser), test=True
        )
        self.original_parser = original_parser()
        self.original_inputs = OriginalInput(
            *parse_args(self.default_args_list, self.original_parser),
            test=True
        )

    def teardown_method(self):
        if os.path.isdir(tmpdir):
            shutil.rmtree(tmpdir)

    def test_original_behaviour(self):
        keys = vars(self.original_inputs).keys()
        for key in keys:
            if key in ["known_args"]:
                continue
            _orig = getattr(self.original_inputs, key)
            _new = getattr(self.my_inputs, key)
            assert _orig == _new

    def test_new_options(self):
        from bilby_nr.bilby_pipe import (
            create_parser, MainInput
        )
        from bilby_pipe.main import parse_args

        self.outdir = tmpdir
        self.default_args_list = [
            "--ini",
            os.path.join(
                os.path.dirname(__file__),
                "test_config_with_additional_arguments.ini"
            ),
            "--outdir",
            self.outdir,
        ]
        self.my_parser = create_parser()
        self.my_inputs = MainInput(
            *parse_args(self.default_args_list, self.my_parser), test=True
        )
        assert sorted(self.my_inputs.waveform_approximant) == sorted([
            "IMRPhenomXPHMST", "IMRPhenomTPHM", "SEOBNRv5PHM"
        ])
        wvf_args = self.my_inputs.get_default_waveform_arguments()
        assert sorted(wvf_args["waveform_approximant_list"]) == sorted([
            "IMRPhenomXPHMST", "IMRPhenomTPHM", "SEOBNRv5PHM"
        ])
        assert wvf_args["waveform_approximant"] == sorted([
            "IMRPhenomXPHMST", "IMRPhenomTPHM", "SEOBNRv5PHM"
        ])[0]


class TestDataAnalysisInput(object):
    def setup_method(self):
        from bilby_nr.bilby_pipe import (
            create_parser, DataAnalysisInput
        )
        from bilby_pipe.data_analysis import (
            create_analysis_parser as original_parser,
            DataAnalysisInput as OriginalInput
        )
        from bilby_pipe.main import parse_args

        self.outdir = tmpdir
        self.default_args_list = [
            "--ini",
            os.path.join(os.path.dirname(__file__), "test_config.ini"),
            "--outdir",
            self.outdir,
        ]
        self.my_parser = create_parser()
        self.my_inputs = DataAnalysisInput(
            *parse_args(self.default_args_list, self.my_parser), test=True
        )
        self.original_parser = original_parser()
        self.original_inputs = OriginalInput(
            *parse_args(self.default_args_list, self.original_parser),
            test=True
        )

    def teardown_method(self):
        if os.path.isdir(tmpdir):
            shutil.rmtree(tmpdir) 

    def test_original_behaviour(self):
        keys = vars(self.original_inputs).keys()
        for key in keys:
            if key in ["known_args"]:
                continue
            _orig = getattr(self.original_inputs, key)
            _new = getattr(self.my_inputs, key)
            assert _orig == _new

    def test_new_options(self):
        from bilby_nr.bilby_pipe import (
            create_parser, DataAnalysisInput
        )
        from bilby_pipe.main import parse_args

        self.outdir = tmpdir
        self.default_args_list = [
            "--ini",
            os.path.join(
                os.path.dirname(__file__),
                "test_config_with_additional_arguments.ini"
            ),
            "--outdir",
            self.outdir,
        ]
        self.my_parser = create_parser()
        self.my_inputs = DataAnalysisInput(
            *parse_args(self.default_args_list, self.my_parser), test=True
        )
        assert sorted(self.my_inputs.waveform_approximant) == sorted([
            "IMRPhenomXPHMST", "IMRPhenomTPHM", "SEOBNRv5PHM"
        ])
        wvf_args = self.my_inputs.get_default_waveform_arguments()
        assert sorted(wvf_args["waveform_approximant_list"]) == sorted([
            "IMRPhenomXPHMST", "IMRPhenomTPHM", "SEOBNRv5PHM"
        ])
        assert wvf_args["waveform_approximant"] == sorted([
            "IMRPhenomXPHMST", "IMRPhenomTPHM", "SEOBNRv5PHM"
        ])[0]



import numpy as np
from bilby.core.utils import create_frequency_series
import pytest


class TestMultiModelModel(object):
    def setup_method(self):
        self.parameters = dict(
            mass_1 = 100,
            mass_2 = 50,
            luminosity_distance = 100,
            a_1 = 0.6,
            tilt_1 = np.pi / 3,
            phi_12 = np.pi / 2,
            a_2 = 0.2,
            tilt_2 = np.pi / 10,
            phi_jl = np.pi,
            theta_jn = np.pi / 3,
            phase = 0.,
        )
        self.waveform_kwargs = dict(
            waveform_approximant="IMRPhenomPv2",
            reference_frequency=50.0,
            minimum_frequency=20.0,
            catch_waveform_errors=True,
            match_interpolant="bilby_nr.match.match_from_interpolant",
            waveform_approximant_list=["IMRPhenomXPHMST", "IMRPhenomTPHM"]
        )
        self.frequency_array = create_frequency_series(2048, 4)

    def test_multi_model_binary_black_hole_no_interp(self):
        from bilby_nr.source import multi_model_binary_black_hole
        _wvf_args = self.waveform_kwargs.copy()
        _wvf_args.pop("match_interpolant")
        self.parameters.update(_wvf_args)
        pols = multi_model_binary_black_hole(
            self.frequency_array, **self.parameters
        )
        assert isinstance(pols, dict)

    def test_multi_model_binary_black_hole_use_best(self):
        from bilby_nr.source import multi_model_binary_black_hole
        for opt in ["True", True, 1]:
            _wvf_args = self.waveform_kwargs.copy()
            _wvf_args["use_best_match"] = opt
            self.parameters.update(_wvf_args)
            pols = multi_model_binary_black_hole(
                self.frequency_array, **self.parameters
            )
            assert isinstance(pols, dict)

    def test_multi_model_binary_black_hole_interp(self):
        from bilby_nr.source import multi_model_binary_black_hole
        self.parameters.update(self.waveform_kwargs)
        pols = multi_model_binary_black_hole(
            self.frequency_array, **self.parameters
        )
        assert isinstance(pols, dict)

    def test_catch_waveform_errors(self):
        from bilby_nr.source import multi_model_binary_black_hole
        _wvf_args = self.waveform_kwargs.copy()
        _wvf_args["catch_waveform_errors"] = False
        _wvf_args["match_interpolant"] = "fake.fake_interpolant"
        with pytest.raises(ValueError):
            self.parameters.update(_wvf_args)
            pols = multi_model_binary_black_hole(
                self.frequency_array, **self.parameters
            )
        _wvf_args = self.waveform_kwargs.copy()
        _wvf_args["catch_waveform_errors"] = True
        _wvf_args["match_interpolant"] = "fake.fake_interpolant"
        self.parameters.update(_wvf_args)
        pols = multi_model_binary_black_hole(
            self.frequency_array, **self.parameters
        )
        assert pols == None

    def test_multi_model_binary_black_hole_fake_interp(self):
        from bilby_nr.source import multi_model_binary_black_hole
        _wvf_args = self.waveform_kwargs.copy()
        _wvf_args["catch_waveform_errors"] = False
        _wvf_args["match_interpolant"] = "fake.fake_interpolant"
        with pytest.raises(ValueError):
            self.parameters.update(_wvf_args)
            pols = multi_model_binary_black_hole(
                self.frequency_array, **self.parameters
            )
        _wvf_args = self.waveform_kwargs.copy()
        _wvf_args["catch_waveform_errors"] = False
        _wvf_args["match_interpolant"] = "bilby_nr.match.no_match_interp"
        with pytest.raises(ValueError):
            self.parameters.update(_wvf_args)
            pols = multi_model_binary_black_hole(
                self.frequency_array, **self.parameters
            )

    def test_multi_model_binary_black_hole_incorrect_weights(self):
        from bilby_nr.source import _multi_model_binary_black_hole
        with pytest.raises(ValueError):
            pols = _multi_model_binary_black_hole(
                [None, None], self.waveform_kwargs["waveform_approximant_list"],
                self.frequency_array, **self.parameters
            )
        with pytest.raises(ValueError):
            pols = _multi_model_binary_black_hole(
                [1.], self.waveform_kwargs["waveform_approximant_list"],
                self.frequency_array, **self.parameters
            )


def test_weights_from_matches():
    from bilby_nr.source import _weights_from_matches
    matches = np.array([0.3, 0.9, 0.8])
    for opt in ["True", True, 1]:
        weights = _weights_from_matches(matches, use_best=opt)
        np.testing.assert_almost_equal(weights, [0., 1., 0.])

    _true = 1 / ((1 - matches)**4)
    _true /= np.sum(_true)
    for opt in ["False", False, 0]:
        weights = _weights_from_matches(matches, use_best=False)
        np.testing.assert_almost_equal(weights, _true)


def test_weights_from_matches_custom_mapping():
    from bilby_nr.source import _weights_from_matches
    matches = np.array([0.3, 0.9, 0.8])
    recommended = _weights_from_matches(matches)
    string_input = _weights_from_matches(
        matches, mapping="1 / ((1 - matches)**4)"
    )
    np.testing.assert_almost_equal(recommended, string_input)

    # test the various failure modes
    with pytest.raises(ValueError):
        string_input = _weights_from_matches(
            matches, mapping="1 / ((1 - variable)**4)"
        )
    with pytest.raises(ValueError):
        string_input = _weights_from_matches(
            matches, mapping="weights = 1 / ((1 - matches)**4)"
        )
    with pytest.raises(ValueError):
        string_input = _weights_from_matches(
            matches, mapping="1 / ((1 - (matches + unknown_variable))**4)"
        )



import numpy as np
import pytest


class TestMatchInterpolants(object):
    def setup_method(self):
        self.parameters = dict(
            mass_1 = 100,
            mass_2 = 50,
            a_1 = 0.6,
            tilt_1 = np.pi / 3,
            phi_12 = np.pi / 2,
            a_2 = 0.2,
            tilt_2 = np.pi / 10,
            phi_jl = np.pi,
            theta_jn = np.pi / 3,
            phase = 0.,
        )
        self.waveform_approximant = "IMRPhenomXPHMST"

    def test_pade_pade_match_interpolant_for_float_input(self):
        from bilby_nr.match import match_from_interpolant

        match = match_from_interpolant(
            self.waveform_approximant,
            self.parameters["mass_1"],
            self.parameters["mass_2"],
            self.parameters["a_1"],
            self.parameters["tilt_1"],
            self.parameters["phi_12"],
            self.parameters["a_2"],
            self.parameters["tilt_2"],
            self.parameters["phi_jl"],
            self.parameters["theta_jn"],
            self.parameters["phase"],
            interp="pade_pade"
        )
        np.testing.assert_almost_equal(match, 0.9926632521327353)

    def test_pade_pade_match_interpolant_for_array_input(self):
        from bilby_nr.match import match_from_interpolant

        _parameters = {key: [item] for key, item in self.parameters.items()}
        match = match_from_interpolant(
            self.waveform_approximant,
            _parameters["mass_1"],
            _parameters["mass_2"],
            _parameters["a_1"],
            _parameters["tilt_1"],
            _parameters["phi_12"],
            _parameters["a_2"],
            _parameters["tilt_2"],
            _parameters["phi_jl"],
            _parameters["theta_jn"],
            _parameters["phase"],
            interp="pade_pade"
        )
        np.testing.assert_almost_equal(match, [0.9926632521327353])

    def test_invalid_interpolant(self):
        from bilby_nr.match import match_from_interpolant

        with pytest.raises(ValueError):
            match = match_from_interpolant(
                self.waveform_approximant,
                self.parameters["mass_1"],
                self.parameters["mass_2"],
                self.parameters["a_1"],
                self.parameters["tilt_1"],
                self.parameters["phi_12"],
                self.parameters["a_2"],
                self.parameters["tilt_2"],
                self.parameters["phi_jl"],
                self.parameters["theta_jn"],
                self.parameters["phase"],
                interp="unknown"
            )



# Licensed under an MIT style license -- see LICENSE.md

import numpy as np

__author__ = ["Charlie Hoy <charlie.hoy@port.ac.uk>"]


def match_from_interpolant(*args, interp="pade_pade"):
    """Return an estimate for the match based on an interpolant fit

    Parameters
    ----------
    *args: list
        The arguments to pass to the interpolant
    interp: str
        The name of the interpolant to use. Default is "pade_pade"

    Returns
    -------
    match: float
        The estimated match
    """
    if interp not in interpolant_map.keys():
        raise ValueError(
            f"Unable to evaluate the mismatch interpolant using {interp}. "
            f"Only allowed interpolants are: {','.join(interpolant_map.keys())}"
        )
    return interpolant_map[interp](*args)


def match_from_pade_pade_interpolant(
    waveform_approximant, mass_1, mass_2, a_1, tilt_1, phi_12, a_2, tilt_2,
    phi_jl, theta_jn, phase,
):
    """Return an estimate for the match based on a Pade-Pade fit, see
    https://www.nature.com/articles/s41550-025-02579-7

    Parameters
    ----------
    waveform_approximant: str
        The waveform approximant to use
    mass_1: float
        The mass of the primary black hole
    mass_2: float
        The mass of the secondary black hole
    a_1: float
        The dimensionless spin magnitude of the primary black hole
    tilt_1: float
        The tilt angle of the primary black hole spin
    phi_12: float
        The difference in azimuthal angle between the two spins
    a_2: float
        The dimensionless spin magnitude of the secondary black hole
    tilt_2: float
        The tilt angle of the secondary black hole spin
    phi_jl: float
        The azimuthal angle of the total angular momentum
    theta_jn: float
        The angle between the total angular momentum and the line of sight
    phase: float
        The phase of the gravitational wave

    Returns
    -------
    match: float
        The estimated match
    """
    from .interp.pade_pade import match_interpolant
    if isinstance(mass_1, (np.ndarray, list)):
        matches = np.zeros(len(mass_1))
        for ii in range(len(mass_1)):
            matches[ii] = match_interpolant(
                waveform_approximant, mass_1[ii], mass_2[ii], a_1[ii],
                tilt_1[ii], phi_12[ii], a_2[ii], tilt_2[ii], phi_jl[ii],
                theta_jn[ii], phase[ii]
            )
        return matches
    return match_interpolant(
        waveform_approximant, mass_1, mass_2, a_1, tilt_1, phi_12, a_2,
        tilt_2, phi_jl, theta_jn, phase
    )


interpolant_map = {
    "pade_pade": match_from_pade_pade_interpolant,
}



import bilby
from bilby.gw import conversion
from gwpy.timeseries import TimeSeries
from bilby_nr.conversion import (
    determine_waveform_approximant_from_likelihood
)
import pytest


def _setup(model="IMRPhenomTPHM"):
    ifo_list = bilby.gw.detector.InterferometerList([])
    for det in ["H1", "L1"]:
        ifo = bilby.gw.detector.get_empty_interferometer(det)
        data = TimeSeries.fetch_open_data(
            det, 1126259462.4 - 2., 1126259462.4 + 2.
        )
        ifo.strain_data.set_from_gwpy_timeseries(data)
        ifo_list.append(ifo)

    waveform_generator = bilby.gw.waveform_generator.WaveformGenerator(
        frequency_domain_source_model=bilby.gw.source.lal_binary_black_hole,
        waveform_arguments={
            "waveform_approximant": model,
            "reference_frequency": 50,
            "minimum_frequency": 20,
        },
    )
    likelihood = bilby.gw.likelihood.GravitationalWaveTransient(
        ifo_list, waveform_generator
    )
    return likelihood, ifo_list, waveform_generator


def test_generate_all_bbh_parameters():
    samples = {
        "mass_1": [36, 35, 37],
        "mass_2": [32, 31, 33],
        "a_1": [0.8, 0.5, 0.7],
        "a_2": [0.5, 0.2, 0.8],
        "tilt_1": [2.4, 1.2, 4.5],
        "tilt_2": [0.8, 0.6, 2.1],
        "phi_12": [5.8, 4.0, 2.1],
        "phi_jl": [0.25, 0.4, 0.1],
        "theta_jn": [2.5, 2.4, 2.6],
        "ra": [2.2, 2.1, 2.3],
        "dec": [-1.22, -1.2, -1.25],
        "geocent_time": [1126259462.408404, 1126259462.408404, 1126259462.408404],
        "phase": [4., 5., 2.],
        "psi": [0.7, 0.5, 0.4],
        "luminosity_distance": [500, 505, 502],
    }
    waveform_approximants = ["IMRPhenomXPHM", "IMRPhenomTPHM", "IMRPhenomPv2"]
    logls = []
    for num in range(len(samples["mass_1"])):
        likelihood, _, _ = _setup(model=waveform_approximants[num])
        _sample = {key: item[num] for key, item in samples.items()}
        logls.append(likelihood.log_likelihood_ratio(_sample))
    samples["log_likelihood"] = logls
    likelihood.waveform_generator.waveform_arguments.update(
        {"waveform_approximant_list": waveform_approximants}
    )
    models = []
    for num in range(len(samples["mass_1"])):
        _sample = {key: item[num] for key, item in samples.items()}
        sample = conversion.generate_all_bbh_parameters(_sample, likelihood)
        models.append(sample["waveform_approximant"])
    assert models == waveform_approximants


def test_determine_waveform_approximant_from_likelihood():
    import bilby
    from gwpy.timeseries import TimeSeries
    from pandas import DataFrame

    likelihood, ifo_list, waveform_generator = _setup()
    sample = {
        "mass_1": 36, "mass_2": 32, "a_1": 0.8, "a_2": 0.7, "tilt_1": 2.4,
        "tilt_2": 0.8, "phi_12": 5.8, "phi_jl": 0.25, "theta_jn": 2.5,
        "ra": 2.2, "dec": -1.22, "geocent_time": 1126259462.408404,
        "phase": 4., "psi": 0.7, "luminosity_distance": 500
    }
    logl = likelihood.log_likelihood_ratio(sample)
    with pytest.raises(ValueError):
        model = determine_waveform_approximant_from_likelihood(
            sample, ["IMRPhenomPv2", "IMRPhenomTPHM", "IMRPhenomXPHM"],
            likelihood
        )
    sample["log_likelihood"] = logl
    model = determine_waveform_approximant_from_likelihood(
        sample, ["IMRPhenomPv2", "IMRPhenomTPHM", "IMRPhenomXPHM"],
        likelihood
    )
    assert model == "IMRPhenomTPHM"

    # try passing a pandas dataframe instead
    _sample = DataFrame(sample, index=[0])
    model = determine_waveform_approximant_from_likelihood(
        _sample, ["IMRPhenomPv2", "IMRPhenomTPHM", "IMRPhenomXPHM"],
        likelihood
    )
    assert model == "IMRPhenomTPHM"

    # try now with marginalization
    priors = bilby.gw.prior.BBHPriorDict()
    for marg in ["distance", "time", "phase"]:
        kwargs = {}
        _sample = sample.copy()
        if marg == "distance":
            kwargs["distance_marginalization"] = True
        elif marg == "time":
            kwargs["time_marginalization"] = True
            _sample["geocent_time"] = likelihood.interferometers.start_time
            _sample["time_jitter"] = 2.
        else:
            kwargs["phase_marginalization"] = True
            _sample["phase"] = 0.
        likelihood = bilby.gw.likelihood.GravitationalWaveTransient(
            ifo_list, waveform_generator, priors=priors, **kwargs
        )
        logl = likelihood.log_likelihood_ratio(_sample)
        _sample["log_likelihood"] = logl
        model = determine_waveform_approximant_from_likelihood(
            _sample, ["IMRPhenomPv2", "IMRPhenomTPHM", "IMRPhenomXPHM"],
            likelihood
        )
        assert model == "IMRPhenomTPHM"



# Licensed under an MIT style license -- see LICENSE.md

import numpy as np
import os
from bilby.gw.conversion import component_masses_to_symmetric_mass_ratio
from ..conversion import chi_par_chi_perp_from_mass_spin

__author__ = ["Charlie Hoy <charlie.hoy@port.ac.uk>"]


def Cijkl(i, j, k, l, fit_coeffs):
    """Return coefficients of the Pade Pade fit as defined in Eq.X of
    https://www.nature.com/articles/s41550-025-02579-7 given data structure
    """
    ind, = np.where(
        (fit_coeffs["i"] == i) & (fit_coeffs["j"] == j) &
        (fit_coeffs["k"] == k) & (fit_coeffs["l"] == l)
    )
    if len(ind) > 1:
        raise ValueError("Duplicated entry")
    return fit_coeffs["Cijkl"][ind[0]]


def ArcTan(x, y):
    """Convert Mathematica definition of ArcTan to numpy definition"""
    return np.arctan2(y, x)


def match_interpolant(
    waveform_approximant, mass_1, mass_2, a_1, tilt_1, phi_12, a_2, tilt_2,
    phi_jl, theta_jn, phase
):
    """Evaluate match interpolant as defined in Eq.X of
    https://www.nature.com/articles/s41550-025-02579-7

    Parameters
    ----------
    waveform_approximant: str
        Name of waveform approximant you wish to evaluate the interpolant for.
        Currently allowed waveform approximants include
        ["IMRPhenomXPHMST", "IMRPhenomTPHM", "SEOBNRv5PHM"]
    mass_1: float
        Detector-frame primary mass of the binary black hole
    mass_2: float
        Detector-frame secondary mass of the binary black hole
    a_1: float
        Magnitude of the spin vector associated with the primary component in
        the binary black hole
    tilt_1: float
        Polar angle of the spin vector associated with the primary component in
        the binary black hole
    phi_12: float
        Azimuthal angle between the primary spin and secondary spin vector
    a_2: float
        Magnitude of the spin vector associated with the secondary component
        in the binary black hole
    tilt_2: float
        Polar angle of the spin vector associated with the secondary component
        in the binary black hole
    phi_jl: float
        Azimuthal angle between the total angular momentum (J) and the orbital
        angular momentum (L)
    theta_jn: float
        Inclination angle of the binary: the angle between the total angular
        momentum (J) and the line
        of sight (N)
    phase: float
        The phase of the binary black hole

    Returns
    -------
    match: float
        An approximate match for a given model
    """
    return 1 - mismatch_interpolant(
        waveform_approximant, mass_1, mass_2, a_1, tilt_1, phi_12, a_2, tilt_2,
        phi_jl, theta_jn, phase
    )


def mismatch_interpolant(
    waveform_approximant, mass_1, mass_2, a_1, tilt_1, phi_12, a_2, tilt_2,
    phi_jl, theta_jn, phase
):
    """Evaluate mismatch interpolant as defined in Eq.X of
    https://www.nature.com/articles/s41550-025-02579-7

    Parameters
    ----------
    waveform_approximant: str
        Name of waveform approximant you wish to evaluate the interpolant for.
        Currently allowed waveform approximants include
        ["IMRPhenomXPHMST", "IMRPhenomTPHM", "SEOBNRv5PHM"]
    mass_1: float
        Detector-frame primary mass of the binary black hole
    mass_2: float
        Detector-frame secondary mass of the binary black hole
    a_1: float
        Magnitude of the spin vector associated with the primary component in
        the binary black hole
    tilt_1: float
        Polar angle of the spin vector associated with the primary component in
        the binary black hole
    phi_12: float
        Azimuthal angle between the primary spin and secondary spin vector
    a_2: float
        Magnitude of the spin vector associated with the secondary component
        in the binary black hole
    tilt_2: float
        Polar angle of the spin vector associated with the secondary component
        in the binary black hole
    phi_jl: float
        Azimuthal angle between the total angular momentum and the orbital
        angular momentum
    theta_jn: float
        Inclination angle of the binary: the angle between the total angular
        momentum and the line of sight
    phase: float
        The phase of the binary black hole

    Returns
    -------
    mismatch: float
        An approximate mismatch for a given model
    """
    allowed_models = ["IMRPhenomXPHMST", "IMRPhenomTPHM", "SEOBNRv5PHM"]
    if waveform_approximant not in allowed_models:
        raise ValueError(
            f"Unable to evaluate interpolant for waveform model "
            f"{waveform_approximant}. Please provide either 'IMRPhenomXPHMST', "
            f"'IMRPhenomTPHM' or 'SEOBNRv5PHM'."
        )
    if mass_1 < mass_2:
        raise ValueError(
            "Secondary mass must be smaller than the primary mass of the binary"
        )

    Mtot = mass_1 + mass_2
    eta = component_masses_to_symmetric_mass_ratio(mass_1, mass_2)
    # interpolant uses f_ref = 10.70629431812844 Hz. We use the same
    # reference frequency here
    chi_par, chi_perp = chi_par_chi_perp_from_mass_spin(
        mass_1, mass_2, a_1, tilt_1, a_2, tilt_2, phi_12, phi_jl, theta_jn,
        phase, 10.70629431812844
    )
    # Read in file containing fitting coefficients
    if waveform_approximant == "IMRPhenomXPHMST":
        identifier = "XPHMST"
    elif waveform_approximant == "IMRPhenomTPHM":
        identifier = "TPHM"
    else:
        identifier = "SEOB"

    filename = os.path.join(
        os.path.dirname(__file__),
        f"NatureAstronomy.XXX.YYY.2025.{identifier}.txt"
    )
        
    with open(filename, "r") as f:
        lines = f.readlines()
        lines = [l.strip() for l in lines]
        original_variables = lines[1].split("\t")
        changed_variables = lines[2].split("\t")

    # Evaluate the fitting co-efficients and return the mismatch
    fit_coeffs = np.genfromtxt(filename, skip_header=3, names=True)
    Sqrt = np.sqrt
    x, y, z, v = original_variables[:4]
    X, Y, Z, V = changed_variables[2:6]

    for var in [x, y, z, v]:
        X = X.replace(var.split("=")[0], f"({var.split('=')[1]})")
        Y = Y.replace(var.split("=")[0], f"({var.split('=')[1]})")
        Z = Z.replace(var.split("=")[0], f"({var.split('=')[1]})")
        V = V.replace(var.split("=")[0], f"({var.split('=')[1]})")

    X = eval(X.split("=")[1])
    Y = eval(Y.split("=")[1])
    Z = eval(Z.split("=")[1])
    V = eval(V.split("=")[1])

    log10_mismatch = np.sum([
        (X)**(i) * (Y)**(j) * np.sum([
            Cijkl(i, j, k, l, fit_coeffs) * (Z)**(k) * (V)**(l)
            for l in np.arange(0, np.max(fit_coeffs["l"]) - 1)
            for k in np.arange(0, np.max(fit_coeffs["k"]) + 1)
        ]) / np.sum([
            np.abs(Cijkl(i, j, k, l + 2, fit_coeffs)) * (Z)**(k) * (V)**(l)
            for l in np.arange(0, np.max(fit_coeffs["l"]) - 1)
            for k in np.arange(0, np.max(fit_coeffs["k"]) + 1)
        ])
        for j in np.arange(0, np.max(fit_coeffs["j"]) + 1)
        for i in np.arange(0, np.max(fit_coeffs["i"]) + 1)
    ])
    mismatch = 10**log10_mismatch
    return mismatch



# Licensed under an MIT style license -- see LICENSE.md

from bilby_pipe.input import Input as _Input
from bilby_pipe.main import MainInput as _MainInput, main
from bilby_pipe.data_analysis import DataAnalysisInput as _DAInput
from bilby_pipe.utils import logger, BilbyPipeError
import inspect

__author__ = ["Charlie Hoy <charlie.hoy@port.ac.uk>"]


class Input(_Input):
    """Superclass of input handlers inherited from bilby_pipe.input.Input
    """
    @property
    def waveform_approximant(self):
        return self._waveform_approximant

    @waveform_approximant.setter
    def waveform_approximant(self, waveform_approximant):
        """Set the waveform approximant.

        If a list of waveform approximants is provided, the code will check if
        the source model is compatible with sampling over multiple models. If it
        is, the code will set the `_sample_multiple_models` flag to True. If
        not, it will raise an error.

        Parameters
        ----------
        waveform_approximant: str or list
            The waveform approximant to use.
        """
        from .utils import convert_waveform_list_from_input
        self._sample_multiple_models = False
        self._waveform_approximant = waveform_approximant
        if waveform_approximant is not None:
            models = convert_waveform_list_from_input(waveform_approximant)
            if isinstance(models, list) and len(models) > 1:
                source_model = self.bilby_frequency_domain_source_model
                source_module = inspect.getmodule(source_model).__name__
                if source_module == "bilby_nr.source":
                    logger.warning(
                        f"A list has been provided for the waveform "
                        f"approximant. Sampling over the models: "
                        f"{','.join(models)}."
                    )
                    self._sample_multiple_models = True
                elif isinstance(models, list):
                    raise BilbyPipeError(
                        f"A waveform approximant list has been provided but "
                        f"a frequency domain source model inside the "
                        f"`bilby_nr` package has not been specified. Unable to "
                        f"sample over multiple models."
                    )
            elif isinstance(models, list):
                # recover the original behaviour
                models = models[0]
            self._waveform_approximant = models

    def get_default_waveform_arguments(self):
        """Return the default waveform arguments.

        If the code is sampling over multiple models, this method will return
        the arguments for the first model in the list, and add the full list
        to the dictionary under the key `waveform_approximant_list`.

        Returns
        -------
        wfa: dict
            The default waveform arguments.
        """
        wfa = super().get_default_waveform_arguments()
        if self._sample_multiple_models:
            # store just the first approximant to prevent issues downstream
            # and add a new waveform argument which is used by the
            # bilby_nr source models
            wfa["waveform_approximant"] = self.waveform_approximant[0]
            wfa["waveform_approximant_list"] = self.waveform_approximant
        return wfa


class MainInput(Input, _MainInput):
    """An object to hold all the inputs to bilby_pipe. Inherited from
    bilby_pipe.main.MainInput
    """
    def __init__(self, *args, **kwargs):
        self.frequency_domain_source_model = args[0].frequency_domain_source_model
        super().__init__(*args, **kwargs)


class DataAnalysisInput(Input, _DAInput):
    """Handles user-input for the data analysis script. Inhertied from
    bilby_pipe.data_analysis.DataAnalysisInput
    """
    def __init__(self, *args, **kwargs):
        self.frequency_domain_source_model = args[0].frequency_domain_source_model
        super().__init__(*args, **kwargs)


def create_parser(top_level=False):
    """Extends the BilbyArgParser for bilby_pipe to include additional
    options for sampling over multiple models

    Parameters
    ----------
    top_level:
        If true, parser is to be used at the top-level with requirement
        checking etc, else it is an internal call and will be ignored.
    """
    from bilby_pipe.parser import create_parser
    parser = create_parser(top_level=top_level)
    parser.set_defaults(
        main_input_class="bilby_nr.bilby_pipe.MainInput",
        analysis_input_class="bilby_nr.bilby_pipe.DataAnalysisInput",
    )
    return parser



# Licensed under an MIT style license -- see LICENSE.md

from bilby_pipe.utils import strip_quotes

__author__ = ["Charlie Hoy <charlie.hoy@port.ac.uk>"]


def convert_waveform_input(string):
    """Convert string inputs into a standard form for the waveform list

    Parameters
    ----------
    string: str
        A string representation to be converted
    """
    if string is None:
        raise ValueError("No input provided")
    if isinstance(string, list):
        string = ",".join(string)
    # Remove square brackets
    string = string.replace("[", "").replace("]", "")
    # Remove added quotes
    string = strip_quotes(string)
    # Replace multiple spaces with a single space
    string = " ".join(string.split())
    # Spaces can be either space or comma in input, convert to comma
    string = string.replace(" ,", ",").replace(", ", ",").replace(" ", ",")
    waveforms = string.split(",")
    waveforms.sort()
    return waveforms


def convert_waveform_list_from_input(waveform_list):
    """Convert a waveform list from the input into a list of strings

    Parameters
    ----------
    waveform_list: str or list
        The input waveform list

    Returns
    -------
    waveform_list: list
        The converted waveform list
    """
    if isinstance(waveform_list, list):
        return waveform_list
    return convert_waveform_input(waveform_list)
